/*
 * Copyright (c) 2012 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *					Anthony Mallet on Wed Mar 14 2012
 */
#include <sys/types.h>
#include <assert.h>
#include <errno.h>
#include <err.h>
#include <fcntl.h>
#ifdef HAVE_FEATURES_H
# include <features.h>
#endif
#include <getopt.h>
#include <pthread.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "csLib.h"
#include "h2devLib.h"
#include "h2initGlob.h"
#include "semLib.h"
#include "taskLib.h"

#include "genom3/c/client.h"
#include "private.h"


/* --- local data ---------------------------------------------------------- */

#define pocolibs_csevent	'e'
#define pocolibs_csend		'q'

static int pocolibs_h2init = 0;
static int pocolibs_tdev = ERROR;
static size_t pocolibs_mboxsize = 0;
static int pocolibs_evpipe[2] = { -1, -1 };
static pthread_t pocolibs_evtid;
static int pocolibs_evtid_init = 0;


static void *	pocolibs_evtask(void *data);


/* --- pocolibs_init ------------------------------------------------------- */

int
pocolibs_init(size_t mboxsize)
{
  char name[H2_DEV_MAX_NAME];
  STATUS s;

  /* general pocolibs init */
  if (!pocolibs_h2init) {
    s = h2initGlob(0);
    if (s != OK) goto error;

    if (pipe(pocolibs_evpipe) ||
        fcntl(pocolibs_evpipe[0], F_SETFL, O_NONBLOCK)) {
      errnoSet(errno);
      goto error;
    }
  }
  pocolibs_h2init++;

  /* mailbox init */
  if (pocolibs_mboxsize > 0 && pocolibs_mboxsize < mboxsize) {
    s = csMboxUpdate(0, mboxsize);
    if (s != OK) goto error;
    pocolibs_mboxsize = mboxsize;
  } else if (pocolibs_mboxsize == 0) {
    snprintf(name, sizeof(name), "cclient-%d", getpid());
    s = csMboxInit(name, 0, mboxsize);
    if (s != OK) goto error;
    pocolibs_mboxsize = mboxsize;
    pocolibs_tdev = taskGetUserData(0);
  }

  /* create event task */
  if (!pocolibs_evtid_init) {
    s = pthread_create(&pocolibs_evtid, NULL, pocolibs_evtask, NULL);
    if (s) goto error;
    pocolibs_evtid_init = 1;
  }

  return 0;

error:
  return errnoGet();
}


/* --- pocolibs_fini ------------------------------------------------------- */

void
pocolibs_fini(void)
{
  int i;

  if (pocolibs_h2init) pocolibs_h2init--;
  if (pocolibs_h2init) return;

  if (pocolibs_evtid_init) {
    h2semGive(H2DEV_TASK_SEM_ID(pocolibs_tdev));
    pthread_join(pocolibs_evtid, NULL);
    pocolibs_evtid_init = 0;
  }

  if (pocolibs_mboxsize) {
    csMboxEnd();
    pocolibs_mboxsize = 0;
    pocolibs_tdev = ERROR;
  }

  for(i=0; i<2; i++) {
    if (pocolibs_evpipe[i] > 0) {
      close(pocolibs_evpipe[i]);
      pocolibs_evpipe[i] = -1;
    }
  }
}


/* --- pocolibs_getopt ----------------------------------------------------- */

/* options */
static const char shortopts_string[] = "+i:";
static struct option longopts_list[] = {
  { "name",		required_argument,	NULL,	'i'},
  { "maxmsg",		required_argument,	NULL,	-'m'},
  { NULL, 0, NULL, 0}
};

int
pocolibs_getopt(int argc, char *argv[], struct pocolibs_options *opts)
{
  int c;
  char *e;

  opts->varmsg_maxsize = 32768;
  opts->instance = NULL;

  /* reset getopt */
#ifdef HAVE_OPTRESET
  optreset = 1;
#endif
#ifdef __GLIBC__
  optind = 0; /* not "1", see gnu manual for getopt(3) ... */
#else
  optind = 1;
#endif

  /* parse command line options */
  opterr = 0;
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'i': opts->instance = optarg; break;
      case -'m': {
        unsigned long s = strtoul(optarg, &e, 0);
        if (*e != 0 || s <= 0) return EINVAL;
        opts->varmsg_maxsize = s * 1024;
        break;
      }

      case '?':
      default:
        c = '?'; break;
    }
  if (argc > optind) c = '?';
  return (c == '?') ? EINVAL : 0;
}


/* --- pocolibs_newrqst ---------------------------------------------------- */

int
pocolibs_newrqst(genom_client h, size_t outsize, int *rqstid)
{
  struct genom_client_request *a;
  int i;
  assert(h);

  *rqstid = -1;
  for(i = h->next; i < h->next + h->nactive; i++)
    if (h->active_rqst[i%h->nactive].status == FREE) {
      *rqstid = i%h->nactive;
      break;
    }
  if (*rqstid < 0) {
    a = realloc(h->active_rqst, (h->nactive+1)*sizeof(*a));
    if (!a) return errno;

    *rqstid = h->nactive;
    h->nactive++;
    h->active_rqst = a;
  }
  h->next = *rqstid + 1;

  h->active_rqst[*rqstid].info = NULL;
  h->active_rqst[*rqstid].rid = -1;
  h->active_rqst[*rqstid].aid = -1;
  h->active_rqst[*rqstid].status = FREE;
  h->active_rqst[*rqstid].error = 0;
  h->active_rqst[*rqstid].output = malloc(outsize);
  h->active_rqst[*rqstid].decode = NULL;
  h->active_rqst[*rqstid].sentcb = NULL;
  h->active_rqst[*rqstid].donecb = NULL;
  h->active_rqst[*rqstid].cb_data = NULL;
  if (!h->active_rqst[*rqstid].output) return errno;

  return 0;
}


/* --- pocolibs_eventfd ---------------------------------------------------- */

int
pocolibs_eventfd(void)
{
  return pocolibs_evpipe[0];
}


/* --- pocolibs_actinfo ---------------------------------------------------- */

const struct genom_service_info *
pocolibs_service_info(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return NULL;
  if (h->active_rqst[rqstid].status == FREE) return NULL;
  return h->active_rqst[rqstid].info;
}


/* --- pocolibs_waitevent -------------------------------------------------- */

int
pocolibs_waitevent(void)
{
  fd_set fdset;
  int s;

  FD_ZERO(&fdset);
  do {
    FD_SET(pocolibs_evpipe[0], &fdset);
    s = select(pocolibs_evpipe[0]+1, &fdset, NULL, NULL, NULL);
  } while(s < 0 && errno == EINTR);

  return s < 0 ? errno : 0;
}


/* --- pocolibs_waitack ---------------------------------------------------- */

int
pocolibs_waitack(genom_client h, int rqstid)
{
  int s;
  if (rqstid < 0 || rqstid >= h->nactive) return h->error = ENOENT;

  while(h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY) {
    s = pocolibs_waitevent();
    if (s) return h->error = s;

    s = pocolibs_doevents(h);
    if (s) return h->error = s;
  }
  return 0;
}


/* --- pocolibs_waitterm --------------------------------------------------- */

int
pocolibs_waitterm(genom_client h, int rqstid)
{
  int s;
  if (rqstid < 0 || rqstid >= h->nactive) return h->error = ENOENT;

  while(h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY ||
        h->active_rqst[rqstid].status == WAITING_FINAL_REPLY) {
    s = pocolibs_waitevent();
    if (s) return h->error = s;

    s = pocolibs_doevents(h);
    if (s) return h->error = s;
  }
  return 0;
}


/* --- pocolibs_done ------------------------------------------------------- */

int
pocolibs_done(genom_client h, int rqstid)
{
  if (rqstid < 0 || rqstid >= h->nactive) return 1;
  if (h->active_rqst[rqstid].status == WAITING_INTERMED_REPLY ||
      h->active_rqst[rqstid].status == WAITING_FINAL_REPLY)
    return 0;
  return 1;
}


/* --- pocolibs_clean ------------------------------------------------------ */

int
pocolibs_clean(genom_client h, int rqstid)
{
  struct genom_client_request *a;
  if (rqstid < 0 || rqstid >= h->nactive) return h->error = ENOENT;

  a = &h->active_rqst[rqstid];
  if (a->status == FREE) return h->error = ENOENT;

  if (a->status == WAITING_INTERMED_REPLY ||
      a->status == WAITING_FINAL_REPLY) {
    h->error = csClientRqstIdFree(h->csid, a->rid);
    if (h->error) return h->error;
  }

  a->status = FREE;
  if (a->output) {
    if (a->info->fini_output) a->info->fini_output(a->output);
    free(a->output);
  }
  return 0;
}


/* --- pocolibs_doevents --------------------------------------------------- */

int
pocolibs_doevents(genom_client h)
{
  static int evcount = 0;

  struct genom_client_request *a;
  ssize_t r;
  char e[16];
  int i, s;

  if (pocolibs_evpipe[0] < 0) return h->error = EBADF;

  do {
    do {
      r = read(pocolibs_evpipe[0], e, sizeof(e));
    } while (r == -1 && errno == EINTR);
    if (r == 0) return h->error = EBADF;

    if (r == -1) {
      if ((errno == EAGAIN || errno == EWOULDBLOCK)) break;
      return h->error = errno;
    }

    for(i=0; i<r; i++) {
      switch(e[i]) {
        case pocolibs_csend:
          return h->error = S_h2devLib_NOT_INITIALIZED;

        case pocolibs_csevent:
          evcount++;
          break;

        default: assert(!"reached unreachable");
      }
    }
  } while(r == sizeof(e));

  if (evcount == 0) return 0;

  for(i = 0; i < h->nactive; i++) {
    a = &h->active_rqst[i];
    if (a->status != WAITING_INTERMED_REPLY &&
        a->status != WAITING_FINAL_REPLY)
      continue;

    s = csClientReplyRcv(
      h->csid, a->rid, NO_BLOCK,
      (char *)&a->aid, sizeof(a->aid), NULL,
      a->output, 0, a->decode);
    if (a->status == s) continue;

    evcount--;
    switch(s) {
      case WAITING_FINAL_REPLY:
        evcount--;
        a->status = s;
        h->error = 0;
        if (a->sentcb)
          (*a->sentcb)(h, i, 0, 0, NULL, a->cb_data);
        break;

      case FINAL_REPLY_OK:
        evcount--;
        a->status = s;
        h->error = 0;
        if (a->donecb)
          (*a->donecb)(h, i, 0, 0, a->output, a->cb_data);
        break;

      default: /* == ERROR */
        h->error = errnoGet();
        a->status = FINAL_REPLY_OK;
        if (a->donecb) {
          if ((h->error & 0xffff0000) != GENOM_COMP_ERR_FLAG)
            (*a->donecb)(h, i, h->error, 0, NULL, a->cb_data);
          else
            (*a->donecb)(h, i, 0, h->error & 0xffff, NULL, a->cb_data);
        }
        break;
    }
  }

  if (evcount < 0) evcount = 0;
  return 0;
}


/* --- pocolibs_evtask ----------------------------------------------------- */

static void *
pocolibs_evtask(void *data)
{
  H2SEM_ID sem;
  char c;

  if (pocolibs_tdev == ERROR) {
    c = pocolibs_csend;
    write(pocolibs_evpipe[1], &c, 1);
    return NULL;
  }

  sem = H2DEV_TASK_SEM_ID(pocolibs_tdev);
  c = pocolibs_csevent;
  while (h2semTake(sem, 0) == TRUE) {
    if (!pocolibs_h2init) break;
    write(pocolibs_evpipe[1], &c, 1);
  }

  c = pocolibs_csend;
  write(pocolibs_evpipe[1], &c, 1);
  return NULL;
}
