<'
# Copyright (c) 2011-2012 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Wed Feb 29 2012
#

if {[llength $argv] != 1} { error "expected arguments: component" }
lassign $argv component

# compute handy shortcuts
set comp [$component name]
set COMP [string toupper [$component name]]
if {[catch { $component version } version]} { set version {}}

lang c
'>
#include "autoconf/acheader.h"

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h>
#include <assert.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <limits.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "portLib.h"
#include "h2devLib.h"
#include "h2initGlob.h"

#include "<"$comp">_control_task.h"


/* --- global data --------------------------------------------------------- */

const char *genom_instance = "<"$comp">";
size_t <"$comp">_varmsg_maxsize = 32768;


/* --- local data ---------------------------------------------------------- */

/* options */
static const char usage_string[] = "\
  -b,--background   daemonize\n\
  -f,--force        bypass multiple component instances detection\n\
  -i,--name=string  set component instance name\n\
     --maxmsg=size  set maximum message size in kB\n\
     --version      display <"$comp"> version number\n\
  -d,--debug        log extra debugging information on stderr\n\
  -h,--help         print usage summary (this text)\n\
";
static const char shortopts_string[] = "+bfi:dh";
static struct option longopts_list[] = {
  { "background",	no_argument,		NULL,	'b'},
  { "force",		no_argument,		NULL,	'f'},
  { "name",		required_argument,	NULL,	'i'},
  { "maxmsg",		required_argument,	NULL,	-'m'},
  { "version",		no_argument,		NULL,	-'v'},
  { "debug",		no_argument,		NULL,	'd'},
  { "help",		no_argument,		NULL,	'h'},
  { NULL, 0, NULL, 0}
};

static int daemonize = 0;
static int force = 0;
static int debug = 0;

static int syncfd[2];		/* daemon synchronization pipe */

static void	genom_<"$comp">_run(void);
static void	genom_<"$comp">_vacuum(const char *pidfile);
static void	usage(FILE *channel, char *argv0);


/* --- main ---------------------------------------------------------------- */

/** Initialization and startup
 */
int
main(int argc, char **argv)
{
  char *e;
  int c;

  /* parse command line options */
  while (
    (c = getopt_long(argc, argv, shortopts_string, longopts_list, NULL)) != -1)
    switch (c) {
      case 0: break;

      case 'b': daemonize++; break;
      case 'f': force++; break;
      case 'd': debug++; break;
      case 'i': genom_instance = optarg; break;

      case -'m': {
        unsigned long s = strtoul(optarg, &e, 0);
        if (*e != 0 || s <= 0) {
          genom_log_warn(0, "bad message size: %s", optarg);
          exit(1);
        }
        <"$comp">_varmsg_maxsize = s * 1024;
        break;
      }

      case -'v': puts("<"$version">"); exit(0); break;
      case 'h':  usage(stdout, argv[0]);  exit(0); break;

      case '?':
      default:
	usage(stderr, argv[0]);
        exit(1);
        break;
    }
  argc -= optind;
  argv += optind;
  if (argc > 0) {
    usage(stderr, argv[0]);
    exit(1);
  }

  /* daemonize */
  if (daemonize) {
    if (pipe(syncfd) < 0) {
      perror("pipe");
      exit(2);
    }

    switch(fork()) {
      case -1:
        perror("cannot fork");
        exit(2);

      case 0: /* child */ {
        int fd;

        close(syncfd[0]);

        if (setsid() == -1) exit(-1);
        if (chdir("/")) /*noop*/;

        fd = open("/dev/null", O_RDWR, 0);
        if (fd != -1) {
          (void)dup2(fd, STDIN_FILENO);
          if (fd > STDERR_FILENO)
            (void)close(fd);
        }
        break;
      }

      default: /* parent */ {
        unsigned char ch;

        close(syncfd[1]);
        ch = '\0';
        while (read(syncfd[0], &ch, sizeof(ch)) < 0 && errno == EINTR)
          ch = '\0';
        _exit(ch == 'x' ? 0:2);
      }
    }
  }

  genom_<"$comp">_run();
  return 0;
}


/* <"[--- ${comp}_run ---------------------------------------------------]"> */

static void
genom_<"$comp">_run(void)
{
  struct utsname uts;
  struct genom_<"$comp">_cntrlstr *cids;
  char pid_file_path[PATH_MAX];
  int pid_file_fd;
  FILE *pid_file;
  char *pid_dir;
  sigset_t sset;
  STATUS s;
  int sig;

  /* block termination signals for all threads */
  sigemptyset(&sset);
  sigaddset(&sset, SIGINT);
  sigaddset(&sset, SIGQUIT);
  sigaddset(&sset, SIGTERM);
  pthread_sigmask(SIG_BLOCK, &sset, NULL);

  /* initialize comLib environnement */
<'
  if {[catch {$component clock-rate} ticks]} {
     set ticks 0
  } else {
     set ticks [expr int(1.0/[$ticks value])]
  }
'>
  s = h2initGlob(<"$ticks">);
  if (s == ERROR) {
    genom_log_warn(1, "h2initGlob error");
    exit(2);
  }

  /* create pid file */
  pid_dir = getenv("H2DEV_DIR");
  if (!pid_dir) pid_dir = getenv("HOME");
  if (!pid_dir) pid_dir = "/tmp";
  assert(pid_dir);

  if (uname(&uts)) {
    genom_log_warn(-1, "uname");
    exit(2);
  }

  snprintf(pid_file_path, PATH_MAX, "%s/.%s.pid-%s",
           pid_dir, genom_instance, uts.nodename);
  pid_file_fd = open(pid_file_path, O_CREAT|O_EXCL|O_WRONLY, 0644);
  if (pid_file_fd < 0) {
    if (force && errno == EEXIST) {
      genom_log_warn(-1, "error creating %s", pid_file_path);
      genom_log_info("forcibly disabling the protection against duplicate"
                     " instances");
      genom_<"$comp">_vacuum(pid_file_path);
      pid_file_fd = open(pid_file_path, O_CREAT|O_EXCL|O_WRONLY, 0644);
    }
    if (pid_file_fd < 0) {
      genom_log_warn(-1, "error creating %s", pid_file_path);
      exit(2);
    }
  }

  pid_file = fdopen(pid_file_fd, "w");
  if (!pid_file) {
    genom_log_warn(-1, "error opening %s", pid_file_path);
    exit(2);
  }
  fprintf(pid_file, "%d\n", getpid());
  fclose(pid_file);

  /* initialize component */
  cids = genom_<"$comp">_init();
  if (!cids) {
    genom_log_warn(0, "initialization failed");
    unlink(pid_file_path);
    exit(2);
  }

  /* signal parent in daemon mode */
  if (daemonize) {
    ssize_t dummy = write(syncfd[1], "x", 1);
    if (dummy) /* silence gcc warnings */;
    close(syncfd[1]);
  }

  /* wait for apocalypse */
  /* For some obscure reasons, Mac OSX SL sigwait can return 0 with sig =0... */
  do { sigwait(&sset, &sig); } while (sig == 0);

  genom_log_info("terminating on %s signal request", strsignal(sig));
  genom_<"$comp">_fini(cids);
  unlink(pid_file_path);
}


/* <"[--- genom_${comp}_vacuum ------------------------------------------]"> */

static void
genom_<"$comp">_vacuum(const char *pidfile)
{
  char n[H2_DEV_MAX_NAME];

  if (unlink(pidfile))
    genom_log_warn(-1, "error unlinking %s", pidfile);

  snprintf(n, sizeof(n), "%s/*", genom_instance);
  genom_log_info("cleaning pocolibs %s device", genom_instance);
  h2devClean(genom_instance);
  genom_log_info("cleaning pocolibs %s devices", n);
  h2devClean(n);
}


/* --- usage --------------------------------------------------------------- */

/** Print usage on a channel.
 *
 * \param[in] channel	Output channel
 * \param[in] argv0	Program invocation name
 */

static void
usage(FILE *channel, char *argv0)
{
  fprintf(channel,
	  "<"[string trim "$comp $version"]"> GenoM component\n\nUsage:\n"
	  "  %1$s [-i name] [--maxmsg size] [-f] [-b] [-d]\n"
	  "  %1$s [-h|--version]\n"
	  "\n%2$s",
	  basename(argv0), usage_string);
}


/* --- log ----------------------------------------------------------------- */

void
genom_log_info(const char *format, ...)
{
  va_list va;

  printf("%s: ", genom_instance);
  va_start(va, format);
  vprintf(format, va);
  va_end(va);
  printf("\n");
}

void
genom_log_warn(int h2error, const char *format, ...)
{
  va_list va;

  fprintf(stderr, "%s: ", genom_instance);
  va_start(va, format);
  vfprintf(stderr, format, va);
  va_end(va);
  if (h2error) {
    fprintf(stderr, ": ");
    if (h2error > 0 ) h2perror(NULL); else perror(NULL);
  } else fprintf(stderr, "\n");
}

void
genom_log_debug(const char *format, ...)
{
  va_list va;
  if (!debug) return;

  fprintf(stderr, "%s: ", genom_instance);
  va_start(va, format);
  vfprintf(stderr, format, va);
  va_end(va);
  fprintf(stderr, "\n");
}
